


def adjacency_list(bigsmiles_: BigSMILES):
    """
    a collection of unordered lists used to represent a finite graph
    Parameters
    ----------
    bigsmiles_

    Returns
    -------

    """
    adjacency_list = []



